DAVIS_SOFTWARE = dict(name='Davis', version=999,
                      author='LaVision GmbH',
                      url='https://www.lavision.de')

DEFAULT_DATASET_LONG_NAMES = {'x': 'x-coordinate',
                              'y': 'y-coordinate',
                              'z': 'z-coordinate',
                              'ix': 'x-pixel-coordinate',
                              'iy': 'y-pixel-coordinate',
                              'reltime': 'reltime'}
