"""const for openpiv module"""
OPENPIV_SOFTWARE = dict(name='openpiv', version=999,
                        url='https://github.com/matthiasprobst/openpiv-python-ext',
                        description="Forked version of openpiv-python with improved I/O and other small improvements.")
